#!/usr/bin/perl

#this script is used for auto SNAP gene prediction
#split the genome into single seq files
#Modify this file if you want to change the way it works
#this script runs with SNAP, so please specify the dir where it is.
#author: gaoshh@big.ac.cn
#v0.1 Nov 22, 2010 basic wrapper for SNAP
#v0.2 Feb 9, 2012 multi-threaded


#modules
use Getopt::Long;
use Bio::Seq;
use Bio::SeqIO;
use File::Spec::Functions qw/rel2abs/;
use File::Basename;
use Thread;


#get options
my %opts;
GetOptions(\%opts,"i:s","o:s","hmm:s","d:s","p:n","aa","tx","xdef:s","name:s","lcmask","plus","minus","gff","ace","quiet","h");

if ((!$opts{i})||(!$opts{hmm})||(!$opts{o})||($opts{h}))
{
	print "
usage: $0 -i <genome> -hmm <HMM file> -o out_pfx [options]
options:
  -h              show help info
  -i <genome>     genome in fasta
  -hmm <hmm>      HMM parameter file
  -o <outpfx>     output prefix
  -d <dir>        working_dir(default is ./)
  -lcmask         treat lowercase as N
  -plus           predict on plus strand only
  -minus          predict on minus strand only
  -gff            output annotation as GFF
  -ace            output annotation as ACED
  -quiet          do not send progress to STDERR(this is forced on, for post-process reasons)
  -aa             create FASTA file of proteins
  -tx             create FASTA file of transcripts
  -xdef <file>    external definitions
  -name <string>  name for the gene [default snap]
  -p num_threads  number of threads to use(defualt: 1)

NOTE: 1.make sure the execution 'snap' is in your PATH
      2.more details about the params, please refer to the SNAP manual
";
	exit(1);
}
$genome=$opts{i};
$hmm=$opts{hmm};
$outpfx=$opts{o};
$aa=$opts{aa};
$tx=$opts{tx};
$xdef=$opts{xdef};
$name=$opts{name};
$lcmask=$opts{lcmask};
$plus=$opts{plus};
$minus=$opts{minus};
$gff=$opts{gff};
$ace=$opts{ace};
$quiet=1;
$wd=($opts{d})?$opts{d}:".";
$np=$opts{p}?$opts{p}:1;

#prepare work dir 	#ver 0.2
$genome=rel2abs($genome);
$wd=rel2abs($wd);
die "[error] $wd not a valid directory!" if ((-e $wd)&&(!-d $wd));
mkdir $wd if (!-e $wd);
system "ln -sf $genome $wd/".basename($genome);
$genome="$wd/".basename($genome);

#split seq
&splitter($genome,$np);

#start snap 
#create np threads
#$seqfile,$partnum,$outpfx,$hmm,$wd,$aa,$tx,$xdef,$name,$lcmask,$plus,$minus,$gff,$ace,$quiet
my %thread;
for $partnum(1..$np){
	my $thread=Thread->new(\&thread_wrapper,"$genome.$partnum",$partnum,$outpfx,$hmm,$wd,$aa,$tx,$xdef,$name,$lcmask,$plus,$minus,$gff,$ace,$quiet);
	$thread{$thread->tid()}=$thread;
	print "Thread $partnum created, tid=".$thread->tid()."\n";
}
#watch threads
my $running=1;
while ($running){
	$running=0;
	foreach $tid(keys %thread){
		my $thread=Thread->object($tid);
		if ($thread->is_running()){
			$running++;
		}else{
			$thread->join();
			delete $thread{$tid};
			print "Thread tid=$tid finished.\n";
		}
	}
	sleep 1;
}
#recycle
system "cat $wd/$outpfx.snap.* > $wd/$outpfx.snap";
system "cat $wd/$outpfx.aa.* > $wd/$outpfx.aa";
system "cat $wd/$outpfx.tx.* > $wd/$outpfx.tx";
print "********************FINISH AUTO GENSCAN********************\n";

#SUB:thread_wrapper	#ver 0.2(2012/2/9)
sub thread_wrapper{
	#threaded genscan wrapper
	#in: $seqfile,$partnum,$smat,$v,$cds,$subopt,$ps,$scale,$outpfx,$wd;
	#ret: is_success?
	my ($seqfile,$partnum,$outpfx,$hmm,$wd,$aa,$tx,$xdef,$name,$lcmask,$plus,$minus,$gff,$ace,$quiet)=@_;
	my $all;
	my $err;
	system "rm -f $wd/$outpfx.snap.$partnum";
	$cmd="/home/xinw/software/snap/snap ";
	$cmd.="$hmm ";		#hmm file
	$cmd.="$seqfile ";		#dna file
	$cmd.="-lcmask " if ($lcmask);
	$cmd.="-plus " if ($plus);
	$cmd.="-minus " if ($minus);
	$cmd.="-gff " if ($gff);
	$cmd.="-ace " if ($ace);
	$cmd.="-quiet " if ($quiet);
	$cmd.="-aa $wd/$outpfx.aa.$partnum " if ($aa);
	$cmd.="-tx $wd/$outpfx.tx.$partnum " if ($tx);
	$cmd.="-xdef $xdef " if ($xdef);
	$cmd.="-name $name " if ($name);
	$cmd.="> $wd/$outpfx.snap.$partnum";		#output file
	$err=system $cmd;
	return $err;
}

#SUB:splitter #ver 0.2(2012/2/9)
sub splitter{
	#fasta seq splitter
	#in: full_path_fasta, part
	#out: no
	my $fasta=shift @_;
	my $part=shift @_;
	my $seqall=`grep -c \">\" $fasta`;
	my $partsize=int($seqall/$part)+1;
	my $seqcount;
	my $partnow;
	my $seqin=Bio::SeqIO->new(-file=>$fasta,-format=>'fasta');
	my $seqout;
	while (my $seq=$seqin->next_seq()){
		$seqcount++;
		if ($seqcount > $partsize*$partnow){
			$partnow++;
			$seqout=Bio::SeqIO->new(-file=>">$fasta.$partnow",-format=>'fasta');
		}
		$seqout->write_seq($seq);
	}
	return 1;
}
