#!/usr/bin/perl

#this script is used for alignment of large query seqs using SPALN

#this script assume that you have spaln set in your PATH
#and ALN_TAB and ALN_DBS have been set in your bash env

#version 1.0 multi-thread by batch and queue
#version 2.0 mutil-thread by average split


#modules
use Getopt::Long;
use Bio::PrimarySeq;
use Bio::SeqIO;
use File::Spec::Functions qw/rel2abs/;
use File::Basename;
use threads;
#use threads::shared;
#use Thread::Queue;

#usage
$usage=<<USAGE;
$0 -p program -q query_file -d database -o out_pfx -t cpu -w work_dir -h
-p program               program can be prot2genome(default),cdna2genome
-q query_file            query in fasta
-d database              database in fasta
-o out_pfx               output prefix
-t cpu                   number of cpus to use
-b batch_size            sequences per batch(default=10)
-w work_dir              working dir for result and temp files

-skipbuild               skip building database index
-skipsplit               skip splitting query
-simpleid                simplify the query ID to >\S+

-- other SPALN params    pass SPALN params like this: -- CQ7 CO0 -M5 [other options]
USAGE

#-preskip                 skip data prepare
#-b batch_size            number of sequences per SPALN batch(default:1000)


#shared declare
#my $fin_count : shared;

#get options
my %opts;
GetOptions(\%opts,"p:s","q:s","d:s","o:s","t:s","b:s","w:s","skipbuild","skipsplit","simpleid","h");
#sanity check
$ALN_TAB=$ENV{'ALN_TAB'};
$ALN_DBS=$ENV{'ALN_DBS'};
if ((!-e $ALN_TAB)||(!-e $ALN_DBS)){
	die "ALN_TAB or ALN_DBS was not set in ENV, check your spaln installation\n";
}
if ((!$opts{q})||(!$opts{d})||($opts{h})){
	die $usage;
}
$program=$opts{p}?$opts{p}:'prot2genome';
$queryf=$opts{q};
$databasef=$opts{d};
$cpu=$opts{t}?$opts{t}:4;
$wd=$opts{w}?$opts{w}:'.';
$batch_size=($opts{b} > 0) ? int($opts{b}) : 10;
$output=$opts{o}?$opts{o}:"$queryf.spaln";
$params=join " ",@ARGV;

die "query file does not exist: $queryf\n" if (!-e $queryf);
die "database file does not exist: $databasef\n" if (!-e $databasef);
die "unsupported program: $program \n" if (($program ne 'prot2genome') && ($program ne 'cdna2genome'));

#prepare params
$wd=rel2abs($wd);
$queryf=rel2abs($queryf);
$output=rel2abs($output);
$databasef=rel2abs($databasef);
$db_basename=substr(basename($databasef),0,12);
$db_basename=~s/\./_/g;

#create workdir
mkdir $wd if (!-e $wd);
for ("split","worker"){
	mkdir ("$wd/$_") if (!-e "$wd/$_");
}
#mkdir "$wd/tmpseq" if (!-e "$wd/tmpseq");

#prepare database
$pwd=`pwd`;
if (!$opts{skipbuild}){
	system "cp $databasef $ALN_DBS/$db_basename.mfa";
	chdir "$ALN_DBS";
	system "make $db_basename.idx";
	system "make $db_basename.bkp" if ($program eq 'prot2genome');
	system "make $db_basename.bkn" if ($program eq 'cdna2genome');
}

#link queryf and split
chdir ($wd);
system("ln -sf $queryf $wd/split/query.fa");
&splitter("$wd/split/query.fa", $cpu) if (!$opts{skipsplit});

#chdir "$wd/tmpseq";

#creat working queue
#$Q_work=Thread::Queue->new();
#$Q_fin=Thread::Queue->new();
#$Q_err = Thread::Queue->new();

#create daemon thread
#$t_daemon=threads->create(\&daemon_thread);
#$t_daemon->detach();

#create working threads and sub working dirs
my @threads;
for $nt(1..$cpu){
	mkdir ("$wd/worker/$nt") if (!-e "$wd/worker/$nt");
	$threads[$nt]=threads->create(\&working_thread, $params, $db_basename, $wd, $batch_size, $nt);
#	$threads[$nt]->detach();
}

#wait for threads to exit
for $nt(1..$cpu){
	$threads[$nt]->join();
}

#combine result;
system("rm -f $wd/output");
system("rm -f $wd/log");
system("rm -f $wd/err");
for $nt(1..$cpu){
	system("cat $wd/worker/$nt/output >> $wd/output");
	system("cat $wd/worker/$nt/log >> $wd/log");
	system("cat $wd/worker/$nt/err >> $wd/err");
}


#split seq into size-fixed batch
#my $batch_id=1;
#my $batch=0;
#my $seqio=Bio::SeqIO->new(-file=>$queryf,-format=>'fasta');
#my $seqbatch=Bio::SeqIO->new(-file=>">$wd/tmpseq/batch.$batch_id",-format=>'fasta');
#while (my $seq=$seqio->next_seq()){
#	next if ($seq->length() < 30);	#skip queries shorter than 30bp/aa
#	if ($batch++ >= $batch_size){
#		$Q_work->enqueue("$batch_id");
#		undef $seqbatch;
#		$batch_id++;
#		$batch=1;
#		$seqbatch=Bio::SeqIO->new(-file=>">$wd/tmpseq/batch.$batch_id",-format=>'fasta');
#	}
#	$seqbatch->write_seq($seq);
#}
#$Q_work->enqueue("$batch_id") if ($batch < $batch_size);
#undef $seqio;
#undef $seqbatch;

#after feeding data
#delete old output if exists
#system "rm -f $output" if (-e $output);
#wait all SPALN batch exit
#while ($Q_work->pending() || $Q_fin->pending()){
#	$fin=$Q_fin->dequeue();
#	$fin_count++;
#	#collect finished result
#	system "cat batch.$fin.output >> $output";
#	system "cat batch.$fin.log >> $output.log";
#	system "cat batch.$fin.err >> $output.err";
#	#clean fininshed batch
#	system "rm if batch.$fin";
#	system "rm -f batch.$fin.output";
#	system "rm -f batch.$fin.log";
#	system "rm -f batch.$fin.err";
#}

##retry once on errors
#while ($Q_err->pending()){
#	$err_id=$Q_err->dequeue();
#	$Q_work->enqueue($err_id);
#}

##wait all SPALN batch exit
#while ($Q_work->pending() || $Q_fin->pending()){
#	$fin=$Q_fin->dequeue();
#	$fin_count++;
#	#collect finished result
#	system "cat batch.$fin.output >> $output";
#	system "cat batch.$fin.log >> $output.log";
#	system "cat batch.$fin.err >> $output.err";
#	#clean fininshed batch
#	system "rm if batch.$fin";
#	system "rm -f batch.$fin.output";
#	system "rm -f batch.$fin.log";
#	system "rm -f batch.$fin.err";
#}

##print out IDs of double errors
#open DBLERR,">","$wd/failed_batch.list";
#while ($Q_err->pending()){
#	$err_id=$Q_err->dequeue();
#		$cmd="spaln ";
#		$cmd.="$params ";
#		$cmd.="-d$db_name ";
#		$cmd.="-obatch.$batch_id.output ";
#		$cmd.="batch.$batch_id ";
#		$cmd.="1>batch.$batch_id.log 2>batch.$batch_id.err ";	
#	print DBLERR "$cmd\n";
#}
#close DBLERR;


#correct output
chdir $wd;
system "mv output output.tmp1";
system "perl -ne 'if (/^#/){print \$block; \$block=\"\";} \$block.=\$_; if (/[^#](#.+)/){\$block=\$1}' output.tmp1 > output.tmp2";
system "perl -ne '\@line=split/\t/; print if (\$line[3]<=\$line[4]);' output.tmp2 > output";
system "rm -f output.tmp*";

#renumber genes
open GFF,"output" || die "who moved my output??\n";
open OUT,">","output.renubmer";
my %hash;
my ($geneID, $mRNAID, $cdsID);
while (<GFF>){
	chomp;
	my ($tid,$source,$type,$start,$end,$score,$dir,$phase,$attrib)=split/\t/;
	if ($type eq 'gene'){
		$geneID++;
		$gID = digit($geneID, 8);
		s/gene\d+/gene$gID/;
	}elsif ($type eq 'mRNA'){
		$mRNAID++;
		$gID = digit($geneID, 8);
		$mID = digit($mRNAID, 8);
		s/mRNA\d+/mRNA$mID/;
		s/gene\d+/gene$gID/;			
	}elsif ($type eq 'cds'){
		$cdsID++;
		$mID = digit($mRNAID, 8);
		$cID = digit($cdsID, 8);
		s/mRNA\d+/mRNA$mID/;
		s/cds\d+/cds$cID/;
	}		
	print OUT "$_\n";
}
close GFF;

sub digit{
	my $num = shift;
	my $digit = shift;
	while (length($num) < $digit){
		$num = "0".$num;
	}
	return $num;
}

system "cat output.renubmer > output";
system "rm -f output.renubmer";

#finishing
system "cp $wd/output $output";
print "Finished auto spaln!\n";
print "Output result can be found in $output\n";
chdir $pwd;

#===================END=================


#working thread
sub working_thread{
	#print @_;

	my $params = shift @_;
	my $db_name = shift @_;
	my $workdir = shift @_;
	my $batchSize = shift @_;
	my $threadID = shift @_;
	
	my $cmd;	#cmdline
	my $t_output = "$workdir/worker/$threadID/output";
	my $t_log = "$workdir/worker/$threadID/log";
	my $t_err = "$workdir/worker/$threadID/err";
	
	#clean old output
	system("rm -f $t_output");
	system("rm -f $t_log");
	system("rm -f $t_err");
	
	my $seqio=Bio::SeqIO->new(-file=>"$workdir/split/query.fa.split.$threadID",-format=>'fasta');
	my $batch_id = 0;
	while (my $seq=$seqio->next_seq()){
		next if ($seq->length() < 30);	#skip queries shorter than 30bp/aa
		$batch_id++;
		my $seqcount = 0;
		open ($seqbatch,">","$workdir/worker/$threadID/query.$threadID.$batch_id");
		
		#print the first one
		$seqcount++;
		print $seqbatch ">".$seq->display_id()."\n";
		print $seqbatch $seq->seq()."\n";
		
		#print the rest
		while (($seqcount < $batchSize) && ($seq=$seqio->next_seq())){
			next if ($seq->length() < 30);	#skip queries shorter than 30bp/aa
			$seqcount++;
			print $seqbatch ">".$seq->display_id()."\n";
			print $seqbatch $seq->seq()."\n";
		}
		close $seqbatch;
		
#		to write seq with bioseq has very bad performance... deprecated!
#		my $seqbatch=Bio::SeqIO->new(-file=>">$workdir/worker/$threadID/query.$threadID.$seqcount",-format=>'fasta');
#		$seqbatch->write_seq($seq);		
#		undef $seqbatch;		
		
		&wrap($params, $db_name, $workdir, $threadID, $batch_id);
		
	}
	undef $seqio;
	print "thread $threadID : finished!\n";
}


sub splitter{
	my $fasta=shift @_;
	my $parts=shift @_;

	for $n(1..$parts){
		open $n,">","$fasta.split.$n" || die;
	}
	
	open FASTA,"<",$fasta || die;
	while(<FASTA>){
		if (/^>/){
			$n=int(rand($parts))+1;
			s/^(>\S+).*/\1/ if ($opts{simpleid});
		}
		print $n $_;
	}

	for $n(1..$parts){
		close OUT;
	}
	close FASTA;
}

sub wrap{
	my ($params, $db_name, $workdir, $threadID, $seqcount) = @_;
	
	my $cmd;	#cmdline
	my $t_output = "$workdir/worker/$threadID/output";
	my $t_log = "$workdir/worker/$threadID/log";
	my $t_err = "$workdir/worker/$threadID/err";

	$cmd="spaln ";
	$cmd.="$params ";
	$cmd.="-d$db_name ";
	$cmd.="-o$workdir/worker/$threadID/query.$threadID.$seqcount.output ";
	$cmd.="$workdir/worker/$threadID/query.$threadID.$seqcount ";
	$cmd.="1>$workdir/worker/$threadID/query.$threadID.$seqcount.log 2>$workdir/worker/$threadID/query.$threadID.$seqcount.err ";
	my $ret = system $cmd;
	if ($ret == 0){
		#success
		#cat result
		system("cat $workdir/worker/$threadID/query.$threadID.$seqcount.output >> $t_output");
		system("cat $workdir/worker/$threadID/query.$threadID.$seqcount.log >> $t_log");
		system("cat $workdir/worker/$threadID/query.$threadID.$seqcount.err >> $t_err");
		#clean
		unlink "$workdir/worker/$threadID/query.$threadID.$seqcount";
		unlink "$workdir/worker/$threadID/query.$threadID.$seqcount.output";
		unlink "$workdir/worker/$threadID/query.$threadID.$seqcount.log";
		unlink "$workdir/worker/$threadID/query.$threadID.$seqcount.err";
	}else{
		#fail
		print "thread $threadID : execution failed on query.$threadID.$seqcount \n";
	}
}


#daemon thread
#sub daemon_thread{
#	while (1){
#		sleep 10;
#		print "Batch Status: submitted:".$Q_work->pending()."\trecycling:".$Q_fin->pending()."\tfinished:".$fin_count."\terror:".$Q_err->pending()."\n";
#	}
#}
