#!/usr/bin/perl -w

=pod
Author: Yong Li
Writing: Mar 27, 2014

This script is for correct ambiguous bases in the genome assembly (for example from AllPaths)
based on read coverage.

Usage Example:
bowtie2-build -f genome.assembly.fa built.genome
bowtie2 -q --rf --phred33 --end-to-end --very-sensitive -I 0 -X 10000 -p 40 -x built.genome -1 mp.R1.fq -2 mp.R2.fq -S mp.sam
samtools view -b -S -o mp.bam mp.sam
samtools sort mp.bam mp.sorted
samtools rmdup -S mp.sorted.bam mp.sorted.uniq.bam
samtools mpileup -f genome.assembly.fa mp.sorted.uniq.bam > variants.vcf
awk '{if ($3 !~ /[ACGT]/) print}' variants.vcf > variants.flt.vcf
correct.ambiguous.pl variants.flt.vcf genome.assembly.fa genome.corrected.fa

=cut

use Bio::SeqIO;
use Getopt::Long;
use List::Util qw/max min/;

my $thisScript = $0; $thisScript = $1 if ($0 =~ /\\([^\\]+)$/ || $0 =~ /\/([^\/]+)$/);

if ( @ARGV != 3 || $ARGV[0] eq "-h" || $ARGV[0] eq "-help" || $ARGV[0] eq "--help" ) {
    print "\nUsage: ./$thisScript in.variants.vcf assembly.fa out.fa [-c minimum_coverage(default: 3)].\n\n";
    exit(1);
}

print "\nRunning $thisScript for $ARGV[0] now ...\n";

my $min_coverage = undef;
GetOptions( 'c|min_coverage' => \$min_coverage);
$min_coverage = 3 if ( !defined $min_coverage );
my $cor_count = 0;
my ($in, @arr, $chr, $pos, $ref, %amount, %qual, $max_base, $id);

my $seqi_object = Bio::SeqIO->new( -file => $ARGV[1], -format => "fasta", verbose => -1 );
my $seqo_object = Bio::SeqIO->new( -file => ">$ARGV[2]", -format => 'fasta', verbose => -1 );
my $seq = $seqi_object->next_seq;
$id = $seq->id;

open( IN,  "<$ARGV[0]" ) or die "Error: Can't open file: $!\n";

while ( defined( $in = <IN> ) ) {
    chomp($in);
	@arr = split /\t/, $in;
	next if $arr[3] < $min_coverage;
	#next if $arr[2] eq 'N';
	$chr = $arr[0];
	$pos = $arr[1];
	$ref = $arr[2];
	$arr[4] =~ s/\^|\$//g;
	$arr[4] =~ tr/acgt/TGCA/;
	$arr[4] =~ s/[^TGCA]//g;
	next if (length $arr[4] != length $arr[5]);
	%amount = ('A' => 0, 'C' => 0, 'G' => 0, 'T' => 0);
	%qual = ('A' => 0, 'C' => 0, 'G' => 0, 'T' => 0);
	for (my $i = 0; $i < length($arr[4]); $i++){
		my $tmp_base = substr($arr[4], $i, 1);
		my $tmp_qual = substr($arr[5], $i, 1);
		$amount{$tmp_base}++;
		$qual{$tmp_base} += ord($tmp_qual);
	}
	my %qual_rev = reverse %qual;
	$max_base = $qual_rev{max(keys %qual_rev)};
	next if $amount{$max_base} < $min_coverage;
	if ($id eq $chr){
		my $sequence = $seq->seq();
		print "$chr_$pos: ref is $ref, real base is " . substr($sequence, $pos - 1, 1) . "\n" if $ref ne substr($sequence, $pos - 1, 1);
		substr($sequence, $pos - 1, 1) = $max_base;
		$seq->seq($sequence);
		$cor_count++;
	}
	else{
		$seqo_object->write_seq($seq);
		while ($seq = $seqi_object->next_seq){
			$id = $seq->id;
			if ($id ne $chr){
				$seqo_object->write_seq($seq);
			}
			else {
				my $sequence = $seq->seq();
				print "$chr_$pos: ref is $ref, real base is " . substr($sequence, $pos - 1, 1) . "\n" if $ref ne substr($sequence, $pos - 1, 1);
				substr($sequence, $pos - 1, 1) = $max_base;
				$seq->seq($sequence);
				$cor_count++;
				last;
			}
		}
	}
}
$seqo_object->write_seq($seq);
while ($seq = $seqi_object->next_seq){
	$seqo_object->write_seq($seq);
}

close(IN);
$seqi_object->close;
$seqo_object->close;
print "\tParameters: " . join(' ', @ARGV) . ".\n";
print "\tMinimum coverage used: $min_coverage.\n";
print "\t$cor_count bases corrected.\n";

print "    ... $thisScript running finished.\n\n";

