#!/usr/bin/perl


#this script is used for...
#extract SEQUENCE from the GFF3 like gene-annotations


#step:
#1. read GFF3 and construct gene structure
#2. if some types of elements is missing, compensate them
#3. read genome sequence
#4. extract the SEQs from gene structure
#5. perform statistics
## author: Xin Wang


$usage=<<_USAGE_;
usage: $0 -i gff3_file -o out_prefix -g genome -s

-i gff3_file             GFF3 gene annotation input file
-o out_prefix            output prefix (default: <gff3_file>)
-g genome                genome file in FASTA format

-s                       perform gene statistics
-f                       add domain/function to extracted sequnece, as provided in GFF3

-gff                     re-generate GFF3 file (include primarySeq.gff3)

-h                       help

This script will generate output files according to your GFF3 element TYPE
For example: if you have 'gene','mRNA','exon' in your GFF3 input, it will generate the following:
out_prefix.gene
out_prefix.mRNA
out_prefix.exon

Meanwhile, 3 files are always created:
out_prefix.transcript    the transcript from exons
out_prefix.coding        the coding sequence from cds
out_prefix.protein       the protein translated from cds


_USAGE_


#modules
use Bio::Seq;
use Getopt::Long;


#get options
my %opts;
GetOptions(\%opts,"i:s","o:s","g:s","s","f","gff","h");


#sanity check
if ((!$opts{i})||(!$opts{g})||($opts{h})){die $usage;}

$INPUT = $opts{i};
$GENOME = $opts{g};
$OUTPUT = ($opts{o})?$opts{o}:$INPUT;
$doSTAT = $opts{s};
$doFUNCTION = $opts{f};
$outGFF = $opts{gff};


die "input gff $INPUT not found" if (!-e $INPUT);
die "genome $GENOME not found" if (!-e $GENOME);

$|=1;

#load genome
print "load genome fasta...";
my $genome = loadGenome($GENOME);
print "done.\n";

#parse gff3
print "parse GFF data...";
my $gff = parseGff($INPUT);
print "done.\n";

#sort GFF
print "sort GFF data...";
sortData($gff);
print "done.\n";

#init gene model
print "initialize gene model...";
my $gmodel = initGeneModel($gff);
print "done.\n";

#fix and compensate gene model
print "fix and compensate gene model...";
compGeneModel($gmodel);
print "done.\n";

#assemble transcript and coding sequences into feature
print "assemble transcript and coding sequences...";
assembleSequence($gmodel, $genome);
print "done.\n";

#generate output
print "generate output...";
$|=0;
writeSequence($gmodel, $genome, $OUTPUT);
$|=1;
print "done.\n";

if ($outGFF){
	print "writing new GFF3 file...";
	writeGFF3($gmodel, $OUTPUT);
	print "done.\n";
}

#statistics
if ($doSTAT){
	print "perform statistics...";
	statGeneModel($gmodel, $genome, $OUTPUT);
	print "done.\n";
}





#SUB : statistics
sub statGeneModel{
	my $gmodel = shift;
	my $genome = shift;
	my $outpfx = shift;
	
	open STAT,">","$outpfx.stat" || die "cannot create stat file: $outpfx.stat\n";
	
	my ($genomeCount, $genomeSize);
	my ($geneCount, $geneTotalSize);
	my ($trCount, $trTotalSize);
	my ($cdsCount, $cdsTotalSize);

	my ($exonCount, $exonTotalSize, $exonGC);
	my ($intronCount, $intronTotalSize, $intronGC);
	my (%donor, %acceptor, %pair);
	my ($utr5Count, $utr5TotalSize, $utr5GC, $geneWithUTR5);
	my ($utr3Count, $utr3TotalSize, $utr3GC, $geneWithUTR3, $geneWithUTR53);
	
	my (%geneAltTR, %geneExon, %geneUTR5, %geneUTR3);
	
	my ($seq,$donor,$acceptor);
	
	foreach my $seq(values %$genome){
		$genomeCount++;
		$genomeSize+=length $seq;
	}
	foreach my $gid(keys %$gmodel){
		$geneCount++;
		$geneTotalSize += $gmodel->{$gid}->{feature}->{length};
		$geneAltTR{scalar(keys %{$gmodel->{$gid}->{transcript}})}++;	#alt splicing of gene
		
		foreach my $tr(keys %{$gmodel->{$gid}->{transcript}}){
			#transcript
			$trCount++;
			$trTotalSize += length($gmodel->{$gid}->{transcript}->{$tr}->{feature}->{transcript});
			
			#CDS
			$cdsCount++;
			$cdsTotalSize += length($gmodel->{$gid}->{transcript}->{$tr}->{feature}->{coding});
			
			#exons
			push @{$geneExon{scalar(@{$gmodel->{$gid}->{transcript}->{$tr}->{exon}})}}, $tr;	#gene exon count
			foreach my $exon(@{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}){
				$exonCount ++;
				$exonTotalSize += $exon->{length};
				$seq = substr($genome->{$exon->{tid}}, $exon->{start}-1, $exon->{length});
				$exonGC += $seq =~s/[GCgc]//g;				
			}
			
			#intron
			foreach my $intron(@{$gmodel->{$gid}->{transcript}->{$tr}->{intron}}){
				$intronCount ++;
				$intronTotalSize += $intron->{length};
				$seq = substr($genome->{$intron->{tid}}, $intron->{start}-1, $intron->{length});
				$seq = revcomp($seq) if ($intron->{dir} eq '-');
				$donor = substr($seq, 0, 2);
				$acceptor = substr($seq, -2, 2);
				$donor{$donor}++;
				$acceptor{$acceptor}++;
				$pair{"$donor..$acceptor"}++;
				$intronGC += $seq =~s/[GCgc]//g;				
			}
			
			
		}
		

	}
	
	
	#PRINT STATISTICS
	print STAT "Statistics of gene annotation $OUTPUT\n";
	print STAT ("=" x 10)."\n";
	print STAT "Genome contig number\t$genomeCount\n";
	print STAT "Genome size (bp)\t$genomeSize\n";
	print STAT "Percent of genome of genes (%)\t".(int(10000 * $geneTotalSize / $genomeSize) / 100)."\n";
	print STAT "\n";
	print STAT "Gene number\t$geneCount\n";
	print STAT "Gene total length\t$geneTotalSize\n";
	print STAT "Gene average length\t".(int(100 * $geneTotalSize / $geneCount) / 100)."\n";
	print STAT "Gene density\t".(1000000 * $geneCount / $genomeSize)."/Mb\n";
	print STAT "\n";
	print STAT "All transcripts\t$trCount\n";
	print STAT "Transcript average length (bp)\t".(int(100 * $trTotalSize / $trCount) / 100)."\n";	
	print STAT "Isoform number\tGene count\n";
	foreach my $count(sort{$a<=>$b} keys %geneAltTR){
		print STAT "$count\t$geneAltTR{$count}\n";
	}
	print STAT "\n";
	print STAT "Average coding sequence length\t".(int(100 * $cdsTotalSize / $cdsCount) / 100)."\n";
	print STAT "\n";
	print STAT "All exons\t$exonCount\n";
	print STAT "Average exon length (bp)\t".(int(100 * $exonTotalSize / $exonCount) / 100)."\n";
	print STAT "Exon GC content (%)\t".(int(10000 * $exonGC / $exonTotalSize) / 100)."\n";
	print STAT "\n";
	print STAT "All introns\t$intronCount\n";
	print STAT "Average intron length (bp)\t".(int(100 * $intronTotalSize / $intronCount) / 100)."\n";
	print STAT "Intron GC content (%)\t".(int(10000 * $intronGC / $intronTotalSize) / 100)."\n";
	print STAT "Splicing pair\tCount\n";
	foreach my $pair(sort keys %pair){
		print STAT "$pair\t$pair{$pair}\n";
	}
	print STAT "\n";
	print STAT "Exon number\tGene count\n";
	foreach my $exon(sort{$a<=>$b} keys %geneExon){
		print STAT "$exon\t";
		print STAT scalar @{$geneExon{$exon}};
		if (scalar @{$geneExon{$exon}} <= 5){
			print STAT "\t";
			print STAT join("\t", @{$geneExon{$exon}});
		}
		print STAT "\n";
	}
	print STAT "\n";
	
	close STAT;
	
}


#SUB : generate new GFF3
sub writeGFF3{
	my $gmodel = shift;
	my $outpfx = shift;
	
	my $prime;
	
	open (GFF3,">","$outpfx.gff3") || die "cannot create new GFF3 output file : $outpfx.gff3\n";
	open (GFF3P,">","$outpfx.primary.gff3") || die "cannot create new primary GFF3 output file : $outpfx.primary.gff3\n";
	
	
	foreach my $gid(sort{
		$gmodel->{$a}->{feature}->{tid} cmp $gmodel->{$b}->{feature}->{tid} || 
		$gmodel->{$a}->{feature}->{start} <=> $gmodel->{$b}->{feature}->{start}
	} keys %$gmodel){
		#PART 1: GENE
		my $gene = $gmodel->{$gid}->{feature};
		print GFF3 join("\t",($gene->{tid},$gene->{source},"gene",$gene->{start},$gene->{end},$gene->{score},$gene->{dir},$gene->{phase},$gene->{attrib}))."\n";
		print GFF3P join("\t",($gene->{tid},$gene->{source},"gene",$gene->{start},$gene->{end},$gene->{score},$gene->{dir},$gene->{phase},$gene->{attrib}))."\n";
		
		$prime = 1;	#open primeseq output
		foreach my $tr(sort{
			$gmodel->{$gid}->{transcript}->{$b}->{feature}->{length} <=>
			$gmodel->{$gid}->{transcript}->{$a}->{feature}->{length}
		} keys %{$gmodel->{$gid}->{transcript}}){
			#PART 2: transcript
			my $transcript = $gmodel->{$gid}->{transcript}->{$tr}->{feature};
			print GFF3 join("\t",($transcript->{tid},$transcript->{source},"mRNA",$transcript->{start},$transcript->{end},$transcript->{score},$transcript->{dir},$transcript->{phase},$transcript->{attrib}))."\n";
			if ($prime){
				print GFF3P join("\t",($transcript->{tid},$transcript->{source},"mRNA",$transcript->{start},$transcript->{end},$transcript->{score},$transcript->{dir},$transcript->{phase},$transcript->{attrib}))."\n";
			}
			
			#PART 3: EXON
			foreach my $exon(@{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}){
				print GFF3 join("\t",($exon->{tid},$exon->{source},"exon",$exon->{start},$exon->{end},$exon->{score},$exon->{dir},$exon->{phase},$exon->{attrib}))."\n";
				if ($prime){
					print GFF3P join("\t",($exon->{tid},$exon->{source},"exon",$exon->{start},$exon->{end},$exon->{score},$exon->{dir},$exon->{phase},$exon->{attrib}))."\n";
				}
			}
			
			#PART 4: intron
			foreach my $intron(@{$gmodel->{$gid}->{transcript}->{$tr}->{intron}}){
				print GFF3 join("\t",($intron->{tid},$intron->{source},"intron",$intron->{start},$intron->{end},$intron->{score},$intron->{dir},$intron->{phase},$intron->{attrib}))."\n";
				if ($prime){
					print GFF3P join("\t",($intron->{tid},$intron->{source},"intron",$intron->{start},$intron->{end},$intron->{score},$intron->{dir},$intron->{phase},$intron->{attrib}))."\n";
				}
			}
			
			#PART 5: UTR-5
			foreach my $utr5(@{$gmodel->{$gid}->{transcript}->{$tr}->{utr5}}){
				print GFF3 join("\t",($utr5->{tid},$utr5->{source},"utr5",$utr5->{start},$utr5->{end},$utr5->{score},$utr5->{dir},$utr5->{phase},$utr5->{attrib}))."\n";
				if ($prime){
					print GFF3P join("\t",($utr5->{tid},$utr5->{source},"utr5",$utr5->{start},$utr5->{end},$utr5->{score},$utr5->{dir},$utr5->{phase},$utr5->{attrib}))."\n";
				}
			}
			
			#PART 6: UTR-3
			foreach my $utr3(@{$gmodel->{$gid}->{transcript}->{$tr}->{utr3}}){
				print GFF3 join("\t",($utr3->{tid},$utr3->{source},"utr3",$utr3->{start},$utr3->{end},$utr3->{score},$utr3->{dir},$utr3->{phase},$utr3->{attrib}))."\n";
				if ($prime){
					print GFF3P join("\t",($utr3->{tid},$utr3->{source},"utr3",$utr3->{start},$utr3->{end},$utr3->{score},$utr3->{dir},$utr3->{phase},$utr3->{attrib}))."\n";
				}
			}
			
			#PATY 7: cds
			foreach my $cds(@{$gmodel->{$gid}->{transcript}->{$tr}->{cds}}){
				print GFF3 join("\t",($cds->{tid},$cds->{source},"CDS",$cds->{start},$cds->{end},$cds->{score},$cds->{dir},$cds->{phase},$cds->{attrib}))."\n";
				if ($prime){
					print GFF3P join("\t",($cds->{tid},$cds->{source},"CDS",$cds->{start},$cds->{end},$cds->{score},$cds->{dir},$cds->{phase},$cds->{attrib}))."\n";
				}
			}
			
			$prime = 0; #shut down primeseq output
		}
	}
	
	close GFF3;
	close GFF3P;
}


#SUB : generate output
sub writeSequence{
	my $gmodel = shift;
	my $genome = shift;
	my $outpfx = shift;
	
	open (GENE,">","$outpfx.gene") || die "cannot create gene output file: $outpfx.gene\n";
	open (EXON,">","$outpfx.exon") || die "cannot create exon output file: $outpfx.exon\n";	
	open (INTRON,">","$outpfx.intron") || die "cannot create intron output file: $outpfx.intron\n";
	open (UTR5,">","$outpfx.utr5") || die "cannot create utr5 output file: $outpfx.utr5\n";
	open (UTR3,">","$outpfx.utr3") || die "cannot create utr5 output file: $outpfx.utr3\n";
	
	open (TR,">","$outpfx.transcript") || die "cannot create transcript output file: $outpfx.transcript\n";	
	open (CDS,">","$outpfx.coding") || die "cannot create coding output file: $outpfx.coding\n";
	open (PROT,">","$outpfx.protein") || die "cannot create protein output file: $outpfx.protein\n";
	
	open (TPRIME,">","$outpfx.transcript.primarySeq") || die "cannot create transcript primary seq output file: $outpfx.transcript.primarySeq\n";	
	open (CPRIME,">","$outpfx.coding.primarySeq") || die "cannot create coding primary seq output file: $outpfx.coding.primarySeq\n";
	open (PPRIME,">","$outpfx.protein.primarySeq") || die "cannot create protein primary seq output file: $outpfx.protein.primarySeq\n";
	open (IPRIME,">","$outpfx.intron.primarySeq") || die "cannot create intron primary seq output file: $outpfx.intron.primarySeq\n";
 	
	my $seq;
	
	foreach my $gid(sort{
		$gmodel->{$a}->{feature}->{tid} cmp $gmodel->{$b}->{feature}->{tid} || 
		$gmodel->{$a}->{feature}->{start} <=> $gmodel->{$b}->{feature}->{start}
	} keys %$gmodel){
		#PART 1: GENE
		my $gene = $gmodel->{$gid}->{feature};
		print GENE ">$gid locus=".$gene->{tid}."(".$gene->{start}.",".$gene->{end}.",".$gene->{dir}."); length=".$gene->{length}.";";
		print GENE " domain=".$gene->{domain}." domain containing;" if ($doFUNCTION && $gene->{domain});
		print GENE " function=".$gene->{function}." like;" if ($doFUNCTION && $gene->{function});
		print GENE " target=".$gene->{target}.";" if ($doFUNCTION && $gene->{target});
		print GENE "\n";
		print GENE substr($genome->{$gene->{tid}}, $gene->{start}-1, $gene->{length})."\n";

		$prime = 1;	#open primeseq output
		foreach my $tr(sort{
				$gmodel->{$gid}->{transcript}->{$b}->{feature}->{length} <=>
				$gmodel->{$gid}->{transcript}->{$a}->{feature}->{length}
			} keys %{$gmodel->{$gid}->{transcript}}){
			#PART 2: transcript
			my $transcript = $gmodel->{$gid}->{transcript}->{$tr}->{feature};
			$seq = $gmodel->{$gid}->{transcript}->{$tr}->{feature}->{transcript};
			print TR ">$tr locus=".$transcript->{tid}."(".$transcript->{start}.",".$transcript->{end}.",".$transcript->{dir}."); length=".length($seq).";";
			print TR " domain=".$transcript->{domain}." domain containing;" if ($doFUNCTION && $transcript->{domain});
			print TR " function=".$transcript->{function}." like;" if ($doFUNCTION && $transcript->{function});
			print TR " target=".$transcript->{target}.";" if ($doFUNCTION && $transcript->{target});			
			print TR "\n";
			print TR $seq."\n";
			if ($prime){
				#2014/7/11 change $tr -> $gid, use gid for primeSeqs
				print TPRIME ">$gid locus=".$transcript->{tid}."(".$transcript->{start}.",".$transcript->{end}.",".$transcript->{dir}."); length=".length($seq).";";
				print TPRIME " domain=".$transcript->{domain}." domain containing;" if ($doFUNCTION && $transcript->{domain});
				print TPRIME " function=".$transcript->{function}." like;" if ($doFUNCTION && $transcript->{function});
				print TPRIME " target=".$transcript->{target}.";" if ($doFUNCTION && $transcript->{target});
				print TPRIME "\n";				
				print TPRIME $seq."\n";
			}
			
			#PART 3: coding
			$seq = $gmodel->{$gid}->{transcript}->{$tr}->{feature}->{coding};
			print CDS ">$tr locus=".$transcript->{tid}."(".$transcript->{cdsStart}.",".$transcript->{cdsEnd}.",".$transcript->{dir}."); length=".length($seq).";";
			print CDS " domain=".$transcript->{domain}." domain containing;" if ($doFUNCTION && $transcript->{domain});
			print CDS " function=".$transcript->{function}." like;" if ($doFUNCTION && $transcript->{function});
			print CDS " target=".$transcript->{target}.";" if ($doFUNCTION && $transcript->{target});
			print CDS "\n";
			print CDS $seq."\n";
			if ($prime){
				#2014/7/11 change $tr -> $gid, use gid for primeSeqs			
				print CPRIME ">$gid locus=".$transcript->{tid}."(".$transcript->{cdsStart}.",".$transcript->{cdsEnd}.",".$transcript->{dir}."); length=".length($seq).";";
				print CPRIME " domain=".$transcript->{domain}." domain containing;" if ($doFUNCTION && $transcript->{domain});
				print CPRIME " function=".$transcript->{function}." like;" if ($doFUNCTION && $transcript->{function});
				print CPRIME " target=".$transcript->{target}.";" if ($doFUNCTION && $transcript->{target});
				print CPRIME "\n";
				print CPRIME $seq."\n";
			}
			
			#PART 4: protein
			$seq = $gmodel->{$gid}->{transcript}->{$tr}->{feature}->{protein};
			print PROT ">$tr locus=".$transcript->{tid}."(".$transcript->{cdsStart}.",".$transcript->{cdsEnd}.",".$transcript->{dir}."); length=".length($seq).";";
			print PROT " domain=".$transcript->{domain}." domain containing;" if ($doFUNCTION && $transcript->{domain});
			print PROT " function=".$transcript->{function}." like;" if ($doFUNCTION && $transcript->{function});
			print PROT " target=".$transcript->{target}.";" if ($doFUNCTION && $transcript->{target});
			print PROT "\n";
			print PROT $seq."\n";			
			if ($prime){
				#2014/7/11 change $tr -> $gid, use gid for primeSeqs
				print PPRIME ">$gid locus=".$transcript->{tid}."(".$transcript->{cdsStart}.",".$transcript->{cdsEnd}.",".$transcript->{dir}."); length=".length($seq).";";
				print PPRIME " domain=".$transcript->{domain}." domain containing;" if ($doFUNCTION && $transcript->{domain});
				print PPRIME " function=".$transcript->{function}." like;" if ($doFUNCTION && $transcript->{function});
				print PPRIME " target=".$transcript->{target}.";" if ($doFUNCTION && $transcript->{target});
				print PPRIME "\n";				
				print PPRIME $seq."\n";					
			}			

			
			#PART 5: EXON
			my $exonID = 0;
			foreach my $exon(@{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}){
				$exonID++;
				print EXON ">$tr.exon$exonID locus=".$exon->{tid}."(".$exon->{start}.",".$exon->{end}.",".$exon->{dir}."); length=".$exon->{length}.";\n";
				$seq = substr($genome->{$exon->{tid}}, $exon->{start}-1, $exon->{length});
				$seq = revcomp($seq) if ($exon->{dir} eq '-');
				print EXON $seq."\n";
			}
			
			#PART 6: INTRON
			my $intronID = 0;
			foreach my $intron(@{$gmodel->{$gid}->{transcript}->{$tr}->{intron}}){
				$intronID++;
				print INTRON ">$tr.intron$intronID locus=".$intron->{tid}."(".$intron->{start}.",".$intron->{end}.",".$intron->{dir}."); length=".$intron->{length}.";\n";
				$seq = substr($genome->{$intron->{tid}}, $intron->{start}-1, $intron->{length});
				$seq = revcomp($seq) if ($intron->{dir} eq '-');
				print INTRON $seq."\n";
				if ($prime){
					print IPRIME ">$gid.intron$intronID locus=".$intron->{tid}."(".$intron->{start}.",".$intron->{end}.",".$intron->{dir}."); length=".$intron->{length}.";\n";
					$seq = substr($genome->{$intron->{tid}}, $intron->{start}-1, $intron->{length});
					$seq = revcomp($seq) if ($intron->{dir} eq '-');
					print IPRIME $seq."\n";				
				}
			}

			$prime = 0; #shut down primeseq output

			#PART 7: UTR-5
			my $utr5ID = 0;
			foreach my $utr5(@{$gmodel->{$gid}->{transcript}->{$tr}->{utr5}}){
				$utr5ID++;
				print UTR5 ">$tr.3utr$utr5ID locus=".$utr5->{tid}."(".$utr5->{start}.",".$utr5->{end}.",".$utr5->{dir}."); length=".$utr5->{length}.";\n";
				$seq = substr($genome->{$utr5->{tid}}, $utr5->{start}-1, $utr5->{length});
				$seq = revcomp($seq) if ($utr5->{dir} eq '-');
				print UTR5 $seq."\n";
			}
			
			#PART 8: UTR-3
			my $utr3ID = 0;
			foreach my $utr3(@{$gmodel->{$gid}->{transcript}->{$tr}->{utr3}}){
				$utr3ID++;
				print UTR3 ">$tr.5utr$utr3ID locus=".$utr3->{tid}."(".$utr3->{start}.",".$utr3->{end}.",".$utr3->{dir}."); length=".$utr3->{length}.";\n";
				$seq = substr($genome->{$utr3->{tid}}, $utr3->{start}-1, $utr3->{length});
				$seq = revcomp($seq) if ($utr3->{dir} eq '-');
				print UTR3 $seq."\n";
			}				
		}
	}
	close GENE;
	close TR;
	close CDS;
	close PROT;
	close EXON;
	close INTRON;
	close UTR5;
	close UTR3;
	close TPRIME;
	close CPRIME;
	close PPRIME;
	close IPRIME;
	
}



#SUB : assemble transcript and coding sequences
sub assembleSequence{
	my $gmodel = shift;
	my $genome = shift;
	
	foreach my $gid(keys %$gmodel){
		foreach my $tr(keys %{$gmodel->{$gid}->{transcript}}){
			#STEP 1: assemble EXON to transcript
			my $seq = "";
			foreach my $exon(@{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}){
				$seq.= substr($genome->{$exon->{tid}}, $exon->{start} - 1, $exon->{length});
			}
			#revcomp if '-'
			if ($gmodel->{$gid}->{transcript}->{$tr}->{feature}->{dir} eq '-'){
				$seq = revcomp($seq);				
			}
			$gmodel->{$gid}->{transcript}->{$tr}->{feature}->{transcript} = $seq;
			
			
			#STEP 2: assemble CDS to coding
			my $seq = "";
			foreach my $cds(@{$gmodel->{$gid}->{transcript}->{$tr}->{cds}}){
				$seq.= substr($genome->{$cds->{tid}}, $cds->{start} - 1, $cds->{length});
			}
			#revcomp if '-'
			if ($gmodel->{$gid}->{transcript}->{$tr}->{feature}->{dir} eq '-'){
				$seq = revcomp($seq);	
			}
			$gmodel->{$gid}->{transcript}->{$tr}->{feature}->{coding} = $seq;

			
			#STEP 3: translate coding to protein
			$gmodel->{$gid}->{transcript}->{$tr}->{feature}->{protein} = translate($seq);
			
		}	
	}
}




#SUB : compensate gene model
sub compGeneModel{
	my $gmodel = shift;
	
	foreach my $gid(keys %$gmodel){
		foreach my $tr(keys %{$gmodel->{$gid}->{transcript}}){
			#STEP 1: CDS is critical, if no CDS, maybe they are indicated by exons, copy exon to CDS
			if (scalar @{$gmodel->{$gid}->{transcript}->{$tr}->{cds}} == 0){
				foreach my $exon(@{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}){
					push @{$gmodel->{$gid}->{transcript}->{$tr}->{cds}}, $exon;
				}
			}
			
			#STEP 2: if no EXON elements, copy CDS to EXON
			if (scalar @{$gmodel->{$gid}->{transcript}->{$tr}->{exon}} == 0){
				foreach my $cds(@{$gmodel->{$gid}->{transcript}->{$tr}->{cds}}){
					push @{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}, $cds;
				}
			}
			
			#STEP 3: if no INTRON element, calculate from EXONs
			if (scalar @{$gmodel->{$gid}->{transcript}->{$tr}->{intron}} == 0){
				my $prevEnd = 0;
				my $intronID = 0;
				foreach my $exon(sort{$a->{start} <=> $b->{start}} @{$gmodel->{$gid}->{transcript}->{$tr}->{exon}}){
					if ($prevEnd > 0){
						$intornID++;
						push @{$gmodel->{$gid}->{transcript}->{$tr}->{intron}}, {
							tid=>$exon->{tid},
							source=>$exon->{source},
							start=>$prevEnd + 1,
							end=>$exon->{start} - 1,
							score=>0,
							dir=>$exon->{dir},					
							attrib=>"ID=$tr.intron$intornID;Parent=$tr;Name=$gid",
							length=>$exon->{start} - $prevEnd - 1
						};
					}
					$prevEnd = $exon->{end};					
				}
			}
			
			#STEP 4: if no TRANSCRIPT element, calculate from initial and terminal EXONs
			if (!defined $gmodel->{$gid}->{transcript}->{$tr}->{feature}){
				my @exons = sort{$a->{start} <=> $b->{start}} @{$gmodel->{$gid}->{transcript}->{$tr}->{exon}};
				my $initial = $exons[0];				
				my $terminal = $exons[-1];				
				$gmodel->{$gid}->{transcript}->{$tr}->{feature} = {
					tid=>$initial->{tid},
					source=>$initial->{source},
					start=>$initial->{start},
					end=>$terminal->{end},
					score=>0,
					dir=>$initial->{dir},					
					attrib=>"ID=$tr;Parent=$gid;Name=$gid",
					length=>$terminal->{end} - $initial->{start} + 1,
					domain=>$initial->{domain},			#2014/6/27  add domain/function info 
					function=>$initial->{function},		#2014/6/27  add domain/function info 
					target=>$initial->{target}			#2014/12/25 add target info
				}				
			}
			
			#STEP 4.1: add cdsStart and cdsEnd to transcript feature
			my @cds = sort{$a->{start} <=> $b->{start}} @{$gmodel->{$gid}->{transcript}->{$tr}->{cds}};
			my $initialCDS = $cds[0];
			my $terminalCDS = $cds[-1];
			$gmodel->{$gid}->{transcript}->{$tr}->{feature}->{cdsStart} = $initialCDS->{start};
			$gmodel->{$gid}->{transcript}->{$tr}->{feature}->{cdsEnd} = $terminalCDS->{end};
			$gmodel->{$gid}->{transcript}->{$tr}->{feature}->{cdsLength} = $terminalCDS->{end} - $initialCDS->{start} + 1;
		}
		
		#STEP 5: if no GENE element, inherit from longest TRANSCRIPT
		if (!defined $gmodel->{$gid}->{feature}){
			@trans = sort{
				$gmodel->{$gid}->{transcript}->{$b}->{feature} 
				<=> $gmodel->{$gid}->{transcript}->{$a}->{feature}
				} keys %{$gmodel->{$gid}->{transcript}};
			my $longest = shift @trans;
			$gmodel->{$gid}->{feature} = {
				tid=>$longest->{tid},
				source=>$longest->{source},
				start=>$longest->{start},
				end=>$longest->{end},
				score=>$longest->{score},
				dir=>$longest->{dir},
				attrib=>"ID=$gid;Name=$gid",
				length=>$longest->{length},
				domain=>$longest->{domain},			#2014/6/27  add domain/function info 
				function=>$longest->{function},		#2014/6/27  add domain/function info 
				target=>$longest->{target}			#2014/12/25 add target info
			}
		}
		
	}
	
	return $gmodel;
}



#SUB : initialize gene model
sub initGeneModel{
	my $gff = shift;
	my %gmodel;
	#gmodel has a data structure of
	#$gmodel{$gid}->{feature,transcript}
	# {feature}->{tid,source,start,end,score,dir,attrib,length}
	# {transcript}->{$transcript_id}
	# 	{$transcript_id}->{feature,exon,intron,cds,utr5,utr3}
	#		{feature}->{tid,source,start,end,score,dir,attrib,length,transcriptLength}
	#		{exon}->[{tid,source,start,end,score,dir,phase,attrib}]
	#		{cds}->[{tid,source,start,end,score,dir,phase,attrib}]
	#		{intron}->[{tid,source,start,end,score,dir,attrib,length,donor,acceptor}]
	#		{utr5}->[{tid,source,start,end,score,dir,attrib}]
	#		{utr3}->[{tid,source,start,end,score,dir,attrib}]
	
	foreach my $tid(keys %$gff){
		foreach my $e(@{$gff->{$tid}}){
			#on their types
			if ($e->{type} =~ /gene/i){
				#gene
				$gmodel{$e->{gid}}->{feature} = {
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1,
					domain=>$e->{domain},		#2014/6/27  add domain/function info 
					function=>$e->{function},	#2014/6/27  add domain/function info
					target=>$e->{target}		#2014/12/25 add target info
				}
			}elsif ($e->{type} =~ /(?:mRNA|transcript)/i){
				#mRNA/transcript
				$gmodel{$e->{gid}}->{transcript}->{$e->{id}}->{feature} = {
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},					
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1,
					domain=>$e->{domain},		#2014/6/27  add domain/function info 
					function=>$e->{function},	#2014/6/27  add domain/function info
					target=>$e->{target}		#2014/12/25 add target info
				}
			}elsif ($e->{type} =~ /(?:exon|Esngl|Einit|Eterm)/i){
				#exon
				push @{$gmodel{$e->{gid}}->{transcript}->{$e->{parent}}->{exon}},{
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},					
					phase=>$e->{phase},
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1,
					domain=>$e->{domain},		#2014/6/27  add domain/function info 
					function=>$e->{function},	#2014/6/27  add domain/function info 
					target=>$e->{target}		#2014/12/25 add target info
				};				
			}elsif ($e->{type} =~ /cds/i){
				#CDS
				push @{$gmodel{$e->{gid}}->{transcript}->{$e->{parent}}->{cds}},{
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},					
					phase=>$e->{phase},
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1,
					domain=>$e->{domain},		#2014/6/27  add domain/function info 
					function=>$e->{function},	#2014/6/27  add domain/function info
					target=>$e->{target}		#2014/12/25 add target info
				};	
			}elsif ($e->{type} =~ /intron/i){
				#intron
				push @{$gmodel{$e->{gid}}->{transcript}->{$e->{parent}}->{intron}},{
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},					
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1
				};	
			}elsif ($e->{type} =~ /(?:five_prime_UTR|utr5|5-utr)/i){
				#5'-UTR
				push @{$gmodel{$e->{gid}}->{transcript}->{$e->{parent}}->{utr5}},{
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},					
					phase=>$e->{phase},
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1
				};
			}elsif ($e->{type} =~ /(?:three_prime_UTR|utr3|3-utr)/i){
				#3'-UTR
				push @{$gmodel{$e->{gid}}->{transcript}->{$e->{parent}}->{utr3}},{
					tid=>$e->{tid},
					source=>$e->{source},
					start=>$e->{start},
					end=>$e->{end},
					score=>$e->{score},
					dir=>$e->{dir},					
					phase=>$e->{phase},
					attrib=>$e->{attrib},
					length=>abs($e->{start} - $e->{end}) + 1
				};
			}
		}		
	}
	
	return \%gmodel;
}



#SUB : read genome into HASH
sub loadGenome{
	my $genome = shift;
	my %genome;
	
	open G,"<",$genome || die "cannot open genome file : $genome\n";
	my $id;
	while (<G>){
		chomp;
		if (/^>(\S+)/){
			$id = $1;
		}else{
			$genome{$id}.=$_;
		}
	}
	return \%genome;
}


#SUB : read GFF file into HASH
sub parseGff{
	my $gff = shift;
	my %parent;
	my %gscore;
	my %glength;
	
	unless (-e $gff) {die "Error: parse_gff: given file $gff doesn't exist\n";}
	my ($tid,$source,$type,$start,$end,$score,$dir,$phase,$attrib);
	
	open my $GFF,"$gff" || die;
	my %hash;
	while (<$GFF>){
		chomp;
		my @temp=split(/\t/,$_);
		if ($#temp != 8) {next;}		
		($tid,$source,$type,$start,$end,$score,$dir,$phase,$attrib)=@temp;
		
		next if ($start > $end); #accept only start < end
		
		my %attr;
		my $id;
		my $pid;
		my %childCount;
		
		#check the gff format
		if ($attrib=~/[^\s;]+=[^;]+/){
			#reading GFF3
			while ($attrib=~/([^\s;]+)=([^;]+)/g){
				$attr{$1}=$2;
			}
			$id=$attr{'ID'};
			$pid=($attr{'Parent'})?$attr{'Parent'}:'';
			$parent{$id}=$pid;
		}elsif ($attrib=~/[\S^;\"]+\s+\"[^;\"]+\"/){					
			#reading GTF2
			while ($attrib=~/([^\s;\"]+)\s+\"([^;\"]+)\"/g){
				$attr{$1}=$2;
			}
			
			$pid = ($attr{'transcript_id'})?$attr{'transcript_id'}:$attr{'gene_id'};
			$childCount{$pid}++;
			$id = "$pid.".$childCount{$pid};			
			$parent{$attr{'transcript_id'}} = $attr{'gene_id'};
			$parent{$id}=$pid;
		}else{
			#reading GFF1
			$pid = $attrib;
			$childCount{$pid}++;
			$id = "$pid.".$childCount{$pid};	
			$parent{$id}=$pid;
		}
		push @{$hash{$tid}},{
			tid=>$tid,
			source=>$source,
			type=>$type,
			start=>$start,
			end=>$end,
			score=>$score,
			dir=>$dir,
			phase=>$phase,
			attrib=>$attrib,
			id => $id,
			parent => $pid,
			length=>abs($start-$end)+1,
			domain=>$attr{'Domain'},		#2014/6/27  add domain info 
			function=>$attr{'Function'},	#2014/6/27  add function info
			target=>$attr{'Target'}			#2014/12/25 add target info, mainly for SPALN
		};			
		
	}
	close $GFF;

	#add gid to elements
	for my $tid(keys %hash){
		for my $e(@{$hash{$tid}}){
			$e->{gid}=getGID($e->{id}, \%parent);
		}
	}
	return \%hash;
}



#SUB: to get gene's GID
sub getGID{
	my $id = shift;
	my $pHash = shift;
	
	if ($pHash->{$id} eq ''){
		return $id;
	}else{
		return getGID($pHash->{$id}, $pHash);			
	}	
}



#SUB: sort GFF by 'start' on scaffold
sub sortData{
	#in HASH_by_parse_gff
	#out HASH_new
	my $data=shift @_;
	my %outdata;
	my ($tid,$e);
	for $tid(keys %$data){
		@{$data->{$tid}}=sort{$a->{start} <=> $b->{start}} @{$data->{$tid}};
	}
	return $data;
}



sub translate{
	#in: cdna seq
	#out: protein seq
	my $seq = shift @_;
	my $seqobj=Bio::Seq->new(-display_id => '',-seq => $seq);	
	my $seqtrans=$seqobj->translate()->seq();
	return $seqtrans;
}

sub revcomp{
	#in: dna
	#out: reverse complementary
	my $seq = shift @_;
	$seq=~tr/ATGCatgc/TACGtacg/;
	$seq=reverse $seq;
	return $seq;
}







